/* Image filter library for sqcam driver by philippe Corbes (04/02/2007)
 *
 * Modified in Gamma.c for use with camsource 0.7.1  20.01.2015 by JK
 * 						juergen.koerner@t-online.de
 *
 * This subroutine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <stdlib.h>
#include <libxml/parser.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include "config.h"

#define MODULE_FILTER
#include "module.h"
#include "image.h"
#include "xmlhelp.h"

#include "gamma_table.h"
#include "image_filter.h"

char *name = "gamma";
char *version = VERSION;


/* defines: */
#define CLIP(color) (unsigned char)(((color)>0xFF)?0xff:(((color)<0)?0:(color)))


/* ---------------------------------------------------------------------------------------- */

struct gammactx {
	int gamma, change, red_gain, green_gain, blue_gain,
	    red_offset, green_offset, blue_offset, enabled,
	    schalt5, schalt4, schalt5s, schalt4s, log, hyst, height, width;
	};

struct 	tm *LokaleZeit;
	time_t Zeit;

int integral_bright = 0;
int new_brightness  = 0;

/* Default values */
int Gamma = 3;
int monat = 0;
int schalt4Offset = 0, schalt5Offset = 0;
int schalt4a = 115, schalt5a = 85;
int Freigabe = 1;



int
auto_gamma(struct image *img, struct image_filter *filter, struct gammaconf *conf)
{
	int old_brightness;
	int diff_bright;
	int schalt4, schalt5, hyst;
	int schalt4s, schalt5s, log;

		{
		if (filter->enabled == 1) {
			old_brightness = new_brightness;
			new_brightness = (int)image_brightness(img);
			schalt4        = conf->schalt4;
			schalt5        = conf->schalt5;
			schalt4s       = conf->schalt4s;
			schalt5s       = conf->schalt5s;
			log	       = conf->log;
			hyst           = conf->hyst;

                        /* Scale 255 (bright) -> 0 (dark)
                           Gamma with hysterese leading over an kind of integralcurve (+/- hysterese)
			   Integralfunction,to count out the switching of the Brightnessautomatic of the camera   */
			diff_bright = (new_brightness - old_brightness);
			if (diff_bright > 0) integral_bright += 1;
			if (diff_bright < 0) integral_bright -= 1;

			/* Switching the gammas */
                        /* Switching to Gamma 5 and check month to correct sun run over a year */
                        if (integral_bright <= (schalt5-hyst)) {
                                if (integral_bright >= 22) {
					Gamma = 5;
					if (Freigabe == 1) {
				    		time(&Zeit);
				    		LokaleZeit = localtime(&Zeit);
				    		/* Select actuell month for year sun running correction and setting the schalt4Offset and schalt5Offset */
				    		monat = LokaleZeit->tm_mon+1;
				    		/* wintertime, no offset needed */
				    		if ((monat == 11) || (monat == 12) || (monat == 1) || (monat == 2)) {
				    			schalt4Offset = 0;
				    			schalt5Offset = 0;
				    			}
				    		/* mars or october time, medium offset needed */
				    		if ((monat == 3) || (monat == 10)) {
				    			schalt4Offset =  (int)((schalt4s-schalt4)/3);
				    			schalt5Offset =  (int)((schalt5s-schalt5)/3);
				    			}
				    		/* april or september time, medium offset needed */
				    		if ((monat == 4) || (monat == 9)) {
				    			schalt4Offset =  (int)(((schalt4s-schalt4)/3)*2);
				    			schalt5Offset =  (int)(((schalt5s-schalt5)/3)*2);
				    			}
				    		/* summertime, high medium offset needed */
				    		if ((monat == 5) || (monat == 6) || (monat == 7) || (monat == 8)) {
				    			schalt4Offset = schalt4s - schalt4;
				    			schalt5Offset = schalt5s - schalt5;
				    			}
				    		/* Adding Year-Offset to schalt4 and schalt5 */
				    		schalt4a = schalt4 + schalt4Offset;
				    		schalt5a = schalt5 + schalt5Offset;
				    		Freigabe = 0;
				    	}
                                }
			}
                        /* Switching to Gamma 4 */
                        if (integral_bright <= (schalt4a-hyst)) {
                                if (integral_bright >= (schalt5a+hyst)) Gamma = 4;
                                }
                        /* Switching to Gamma 3 */
			if (integral_bright <  256 ) {
				if (integral_bright >= (schalt4a+hyst)) Gamma = 3;
				}

			/* Regarding the borders */
			if (diff_bright > 10) integral_bright = integral_bright + diff_bright;
			if (integral_bright < 0) integral_bright = 0;

			/*  Switching to Gamma 6 just before reaching darkness depending on Image_Brightness */
			if (new_brightness < 20) Gamma = 6;

			/* Before reaching the total darkness hold to Gamma 6, then switch to Gamma 3 for noise reduction */
			if (new_brightness <= 1) {
				Gamma           = 3;
				integral_bright = 0;
				Freigabe        = 1;
				}

			/* Activate in case of testing: */
			if (log == 1) {
				fprintf(stdout, "B: %i  I: %i  G: %i s5a: %i s4a: %i fg: %i\n", new_brightness, integral_bright, Gamma, schalt5a, schalt4a, Freigabe);
			}
			/* Writing to camsource.log file. */


			filter->Gamma = Gamma;
			}
		}

	return filter->Gamma;
}


static
struct gammactx *
ctx_init(xmlNodePtr node)
{
	struct gammactx *ctx;
	char *cont;

	ctx = malloc(sizeof(*ctx));
	memset(ctx, 0, sizeof(*ctx));

	for (node = node->xml_children; node; node = node->next)
	{
		if (xml_isnode(node, "gamma"))
		{
			ctx->gamma = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "change"))
		{
			cont = xml_getcontent(node);
			if (cont	&& (!strcmp(cont, "yes")
					|| !strcmp(cont, "on")
					|| !strcmp(cont, "1")))
			{
			ctx->change = 1;
			}
			else
			{
			ctx->change = 0;
			}
		}
                if (xml_isnode(node, "enable"))
                {
                        cont = xml_getcontent(node);
                        if (cont        && (!strcmp(cont, "yes")
                                        || !strcmp(cont, "on")
                                        || !strcmp(cont, "1")))
                        {
                        ctx->enabled = 1;
                        }
                        else
                        {
                        ctx->enabled = 0;
                        }
                }
		if (xml_isnode(node, "red_gain"))
		{
			ctx->red_gain = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "green_gain"))
		{
			ctx->green_gain = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "blue_gain"))
		{
			ctx->blue_gain = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "red_offset"))
		{
			ctx->red_offset = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "green_offset"))
		{
			ctx->green_offset = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "blue_offset"))
		{
			ctx->blue_offset = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "schalt4"))
		{
			ctx->schalt4 = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "schalt5"))
		{
			ctx->schalt5 = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "schalt4_Sommer"))
	        {
			ctx->schalt4s = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "schalt5_Sommer"))
		{
			ctx->schalt5s = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "log"))
		{
			cont = xml_getcontent(node);
			if (cont	&& (!strcmp(cont, "yes")
					|| !strcmp(cont, "on")
					|| !strcmp(cont, "1")))
			{
			ctx->log = 1;
			}
			else
			{
			ctx->log = 0;
			}
		}
		if (xml_isnode(node, "hyst"))
		{
			ctx->hyst = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "width"))
		{
			ctx->height = xml_atoi(node, 0);
		}
		if (xml_isnode(node, "height"))
		{
			ctx->width = xml_atoi(node, 0);
		}
	}
	return ctx;
}


/*    Gamma, RGB Gain and Offset    */

/*
 *      GammaTable[0][n] -> 2.2
 *      GammaTable[1][n] -> 1.7
 *      GammaTable[2][n] -> 1.45
 *      GammaTable[3][n] -> 1
 *      GammaTable[4][n] -> 0.6896
 *      GammaTable[5][n] -> 0.5882
 *      GammaTable[6][n] -> 0.4545
*/


void
image_filter_update(struct image_filter *filter)
{
	unsigned char *red   = filter->red;
	unsigned char *green = filter->green;
	unsigned char *blue  = filter->blue;

	int i;


	if (filter->change != 1)  {   /*  Use Default values in case of change = 0 */

		filter->Gamma		=   3;
		filter->GainRed		= 256;  // => GainRed   = GainRed   * 1
		filter->GainGreen	= 256;  // => GainGreen = GainGreen * 1
		filter->GainBlue	= 256;  // => GainBlue  = GainBlue  * 1
		filter->OffsetRed	=   0;
		filter->OffsetGreen	=   0;
		filter->OffsetBlue	=   0;
		}

	/* Gamma setting change compute all case */
	memcpy(red,   GammaTable[filter->Gamma], 256);
	memcpy(green, GammaTable[filter->Gamma], 256);
	memcpy(blue,  GammaTable[filter->Gamma], 256);
	for (i = 0; i < 256; i++) {
		red[i]   = CLIP(((red[i]   + filter->OffsetRed)   * filter->GainRed) >> 8);
		green[i] = CLIP(((green[i] + filter->OffsetGreen) * filter->GainGreen) >> 8);
		blue[i]  = CLIP(((blue[i]  + filter->OffsetBlue)  * filter->GainBlue) >> 8);
		}
	*filter->red   = *red;
     	*filter->green = *green;
	*filter->blue  = *blue;
}


void
image_filter_apply(unsigned char *src_region, unsigned char *dest_region, struct image_filter *filter, struct gammaconf *conf)
{
        int w, h;

        int sel_width;
        int sel_height;

        unsigned char *src,  *s;
	unsigned char *dest, *d;

        sel_width  = conf->width;
        sel_height = conf->height;

	h    = sel_height;
	src  = src_region;
	dest = dest_region;

	while (h--)
	{
		w = sel_width;
		s = src;
		d = dest;
		while (w--)
		{
	                d[0] = filter->red[s[0]];
                	d[1] = filter->green[s[1]];
        	        d[2] = filter->blue[s[2]];

			s += 3;
			d += 3;
		}

	src  += 3*sel_width;
	dest += 3*sel_width;
	}
}


/* ----------------------------------------------------------------------- */


int
filter(struct image *img, xmlNodePtr node, void **instctx)
{
	struct image_filter *filter;
	struct gammaconf *conf;

	struct image *work;
	struct gammactx *ctx;

	unsigned char *r, *w;


	filter = malloc(sizeof(*filter));
	conf   = malloc(sizeof(*conf));

	if (!*instctx) {
		ctx = ctx_init(node);
		*instctx = ctx;
	}
	else
		ctx = *instctx;


	filter->change      = ctx->change;
	filter->enabled     = ctx->enabled;

	filter->Gamma       = ctx->gamma;
	filter->OffsetRed   = ctx->red_offset;
	filter->OffsetGreen = ctx->green_offset;
	filter->OffsetBlue  = ctx->blue_offset;
	filter->GainRed     = ctx->red_gain;
	filter->GainGreen   = ctx->green_gain;
	filter->GainBlue    = ctx->blue_gain;

	conf->schalt4       = ctx->schalt4;
	conf->schalt5       = ctx->schalt5;
	conf->schalt4s      = ctx->schalt4s;
	conf->schalt5s      = ctx->schalt5s;
	conf->log	    = ctx->log;
	conf->hyst	    = ctx->hyst;

	conf->height        = ctx->height;
	conf->width	    = ctx->width;

	work = img;

	r = img->buf;
	w = work->buf;

	auto_gamma(img, filter, conf);
	image_filter_update(filter);
	image_filter_apply(r, w, filter, conf);

	free(conf);
	free(filter);

	return 0;
}

//   ---------------- End of Gamma.c ------------------
